@@ -18,7 +18,7 @@
 */
 /* Implementation of the Message Payload used as channel messages and
    private messages. */
-/* $Id: silcmessage.c,v 1.2.2.6 2002/12/04 18:08:21 priikone Exp $ */
+/* $Id: silcmessage.c,v 1.2.2.7 2002/12/14 13:18:32 priikone Exp $ */
 
 #include "silcincludes.h"
 #include "silcmessage.h"
@@ -544,7 +544,7 @@ silc_message_signed_payload_encode(const unsigned char *message_payload,
 {
   SilcBuffer buffer, sign;
   SilcPKCS pkcs;
-  unsigned char auth_data[2048];
+  unsigned char auth_data[2048 + 1];
   SilcUInt32 auth_len;
   unsigned char *pk = NULL;
   SilcUInt32 pk_len = 0;
